import numpy as np


class Environment:
    """
    Base class for environments. Environments can have conditions that change over time.
    Conditions are objects that have an update method.

    Parameters
    ----------
    dimensions : int, optional
        The number of dimensions of the environment. Default is 3.
    size : float, optional
        The size of the environment. Default is 100.0.

    Attributes
    ----------
    conditions : list
        List of conditions in the environment.
    dimensions : int
        The number of dimensions of the environment.
    size : float
        The size of the environment.

    Methods
    -------
    update()
        Update the environment.
    add_condition(condition)
        Add a condition to the environment
    """
    def __init__(self, dimensions=3, size=100.0):
        self.conditions = []
        self.dimensions = dimensions
        self.size = size

    def update(self):
        """
        Update the environment.
        This method calls the update method of all conditions in the environment.
        """
        for condition in self.conditions:
            condition.update()

    def add_condition(self, condition):
        """
        Add a condition to the environment.

        Parameters
        ----------
        condition : object
            The condition to add to the environment.
        """
        self.conditions.append(condition)


class GradientEnvironment(Environment):
    """
    Environment with a gradient that can be used by agents to navigate.
    It is a subclass of the base Environment class.

    Parameters
    ----------
    gradient_func : function
        The function that calculates the gradient at a given position.
    dimensions : int, optional
        The number of dimensions of the environment. Default is 3.
    size : float, optional
        The size of the environment. Default is 100.0.

    Attributes
    ----------
    conditions : list
        List of conditions in the environment.
    dimensions : int, optional
        The number of dimensions of the environment.
    size : float, optional
        The size of the environment.
    gradient_func : function
        The function that calculates the gradient at a given position.

    Methods
    -------
    update()
        Update the environment.
    add_condition(condition)
        Add a condition to the environment
    get_gradient(position)
        Get the gradient at a given position.
    """
    def __init__(self, gradient_func, dimensions=3, size=100.0):
        super().__init__(dimensions, size)
        self.gradient_func = gradient_func

    def get_gradient(self, position):
        """
        Get the gradient at a given position.

        Parameters
        ----------
        position : array-like
            The position at which to calculate the gradient.

        Returns
        -------
        array-like
            The gradient at the given position.
        """
        return self.gradient_func(position)


class TemperatureEnvironment(Environment):
    """
    Environment with a temperature that can change over time.
    It is a subclass of the base Environment class.

    Parameters
    ----------
    initial_temperature : float
        The initial temperature of the environment.
    temperature_range : tuple, optional
        The range of temperatures that the environment can have. Default is None.
    dimensions : int, optional
        The number of dimensions of the environment. Default is 3.
    size : float, optional
        The size of the environment. Default is 100.0.

    Attributes
    ----------
    conditions : list
        List of conditions in the environment.
    dimensions : int, optional
        The number of dimensions of the environment. Default is 3.
    size : float, optional
        The size of the environment. Default is 100.0.
    temperature : float
        The temperature of the environment.
    temperature_range : tuple
        The range of temperatures that the environment can have (optional).

    Methods
    -------
    update()
        Update the environment.
    add_condition(condition)
        Add a condition to the environment
    get_temperature()
        Get the temperature of the environment.
    set_temperature(temperature)
        Set the temperature of the environment.
    """
    def __init__(self, initial_temperature, temperature_range=None, dimensions=3, size=100.0):
        super().__init__(dimensions, size)
        self.temperature = initial_temperature
        self.temperature_range = temperature_range

    def update(self):
        """
        Update the environment.
        This method updates the temperature of the environment if a temperature range is specified.
        """
        if self.temperature_range is not None:
            self.temperature = np.random.uniform(*self.temperature_range)
        super().update()

    def get_temperature(self):
        """
        Get the temperature of the environment.

        Returns
        -------
        float
            The temperature of the environment.
        """
        return self.temperature

    def set_temperature(self, temperature):
        """
        Set the temperature of the environment.

        Parameters
        ----------
        temperature : float
            The temperature to set
        """
        self.temperature = temperature


class StochasticEnvironment(Environment):
    """
    Environment with a stochastic noise level that can change over time.
    It is a subclass of the base Environment class.

    Parameters
    ----------
    noise_level : float
        The noise level of the environment
    dimensions : int, optional
        The number of dimensions of the environment. Default is 3.
    size : float, optional
        The size of the environment. Default is 100.0.

    Attributes
    ----------
    conditions : list
        List of conditions in the environment.
    dimensions : int, optional
        The number of dimensions of the environment.
    size : float, optional
        The size of the environment.
    noise_level : float
        The noise level of the environment.

    Methods
    -------
    update()
        Update the environment.
    add_condition(condition)
        Add a condition to the environment
    get_noise()
        Get the noise level of the environment.
    """
    def __init__(self, noise_level, dimensions=3, size=100.0):
        super().__init__(dimensions, size)
        self.noise_level = noise_level

    def get_noise(self):
        """
        Get the noise level of the environment.

        Returns
        -------
        float
            The noise level of the environment.
        """
        return np.random.normal(0, self.noise_level)


class ChemicalGradientEnvironment(Environment):
    """
    Environment with a chemical gradient that can be used by agents to navigate.
    It is a subclass of the base Environment class.

    Parameters
    ----------
    gradient_func : function
        The function that calculates the gradient at a given position. Default is np.linalg.norm.
    dimensions : int, optional
        The number of dimensions of the environment. Default is 3.
    size : float, optional
        The size of the environment. Default is 100.0.

    Attributes
    ----------
    conditions : list
        List of conditions in the environment.
    dimensions : int, optional
        The number of dimensions of the environment.
    size : float, optional
        The size of the environment.
    gradient_func : function
        The function that calculates the gradient at a given position.

    Methods
    -------
    update()
        Update the environment.
    add_condition(condition)
        Add a condition to the environment
    get_concentration(position)
        Get the concentration of the chemical at a given position.
    """
    def __init__(self, gradient_func=np.linalg.norm, dimensions=3, size=100.0):
        super().__init__(dimensions, size)
        self.gradient_func = gradient_func

    def get_concentration(self, position):
        """
        Get the concentration of the chemical at a given position.

        Parameters
        ----------
        position : array-like
            The position at which to calculate the concentration.
        """
        return self.gradient_func(position)


class ElectricFieldEnvironment(Environment):
    """
    Environment with an electric field that can be used by agents to navigate.
    It is a subclass of the base Environment class.

    Parameters
    ----------
    field_strength : float
        The strength of the electric field.
    dimensions : int, optional
        The number of dimensions of the environment. Default is 3.
    size : float, optional
        The size of the environment. Default is 100.0.

    Attributes
    ----------
    conditions : list
        List of conditions in the environment.
    dimensions : int, optional
        The number of dimensions of the environment.
    size : float, optional
        The size of the environment.
    field_strength : float
        The strength of the electric field.

    Methods
    -------
    update()
        Update the environment.
    add_condition(condition)
        Add a condition to the environment
    get_field_effect(position)
        Get the effect of the electric field at a given position.
    """
    def __init__(self, field_strength, dimensions=3, size=100.0):
        super().__init__(dimensions, size)
        self.field_strength = field_strength

    def get_field_effect(self, position):
        """
        Get the effect of the electric field at a given position.

        Parameters
        ----------
        position : array-like
            The position at which to calculate the field effect.
        """
        return self.field_strength * position  # May be overridden in subclasses for more complex behavior



class Agent:
    """
    Base class for all agents in the simulation (not to be instantiated directly). It represents a single entity in the
    environment and is used as an interface for different types of agents (e.g., cells, molecules).
    The agent can have multiple modules that define its behavior.

    Parameters
    ----------
    position : tuple
        The position of the agent in the environment.

    Attributes
    ----------
    position : {tuple, int}
        The position of the agent in the environment.
    modules : list
        List of modules that define the behavior of the agent.

    Methods
    -------
    add_module(module)
        Add a module to the agent.
    update()
        Update the agent's state based on its modules.
    """

    def __init__(self, position):
        self.position = position
        self.modules = []

    def add_module(self, module):
        """
        Add a module to the agent.

        Parameters
        ----------
        module : BaseModule
            The module to add to the agent.
        """
        self.modules.append(module)

    def update(self):
        """
        Update the agent's state based on its modules.
        This method should be called at each time step to update the agent's state based on its modules.
        """
        for module in self.modules:
            module.run(self)


class Cell(Agent):
    """
    Base class for all cell agents in the simulation (typically not instantiated directly). It represents a single cell
    in the environment and extends the Agent class with additional functionality specific to cells.
    It is a subclass of the Agent class.

    Parameters
    ----------
    position : tuple
        The position of the cell in the environment.
    input_data_func : function, optional
        The input data function for the cell. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function
        The input data function for the cell.

    Methods
    -------
    add_module(module)
        Add a module to the cell.
    update(historical_value=None)
        Update the cell's state based on its modules and historical value.
    get_history()
        Get the historical values of the cell.
    default_input_data_func()
        Default input data function for the cell.
    get_input_data()
        Get the input data for the cell.
    """

    def __init__(self, position, input_data_func=None):
        super().__init__(position)
        self.history = []
        self.input_data_func = input_data_func if input_data_func is not None \
            else lambda _: (_ for _ in ()).throw(NotImplementedError("Subclasses should implement this method!"))

    def update(self, historical_value=None):
        """
        Update the cell's state based on its modules and historical value.

        Parameters
        ----------
        historical_value : any, optional
            The historical value to append to the cell's history. Default is None.
        """
        super().update()
        if historical_value is not None:
            self.history.append(historical_value)

    def get_history(self):
        """
        Get the historical values of the cell.

        Returns
        -------
        list
            List of historical values for the cell.
        """
        return self.history

    def get_input_data(self, _=None):
        """
        Get the input data for the neuron.

        Returns
        -------
        list
            List of input data values.
        """
        return self.input_data_func(_)


class Synapse:
    """
    Base class for synapses connecting neurons in the simulation. It represents a connection between two neurons and
    is used to transmit signals between them.

    Parameters
    ----------
    pre_neuron : SpikingNeuronCell
        The pre-synaptic neuron.
    post_neuron : SpikingNeuronCell
        The post-synaptic neuron.
    weight : float, optional
        The weight of the synapse. Default is 0.5.

    Attributes
    ----------
    pre_neuron : SpikingNeuronCell
        The pre-synaptic neuron.
    post_neuron : SpikingNeuronCell
        The post-synaptic neuron.
    weight : float
        The weight of the synapse.
    """

    def __init__(self, pre_neuron, post_neuron, weight=0.5):
        self.pre_neuron = pre_neuron
        self.post_neuron = post_neuron
        self.weight = weight

    def transmit(self):
        """
        Transmit a signal from the pre-synaptic neuron to the post-synaptic neuron.
        This method is called when the pre-synaptic neuron fires an action potential.
        The post-synaptic neuron's membrane potential is increased by the weight of the synapse.
        """
        if self.pre_neuron.membrane_potential >= self.pre_neuron.threshold:
            self.post_neuron.membrane_potential += self.weight


class Organoid:
    """
    Base class for all organoids in the simulation (typically not instantiated directly). It represents a collection of
    agents in the environment and is used to simulate the behavior of a group of agents.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid exists.

    Attributes
    ----------
    environment : Environment
        The environment in which the organoid exists.
    agents : list
        List of agents (e.g., cells) in the organoid.

    Methods
    -------
    add_agent(agent)
        Add an agent to the organoid.
    get_cells()
        Get the list of agents in the organoid.
    """

    def __init__(self, environment):
        self.environment = environment
        self.agents = []

    def add_agent(self, agent):
        """
        Add an agent to the organoid.

        Parameters
        ----------
        agent : Agent
            The agent to add to the organoid.
        """
        self.agents.append(agent)

    def get_cells(self):
        """
        Get the list of agents in the organoid.

        Returns
        -------
        list
            List of agents (e.g., cells) in the organoid.
        """
        return self.agents

    def plot_organoid(self, filename=None, show_properties=True, dpi=300, truncate_cells=10):
        """
        Plot the structure of the organoid using Graphviz.

        Parameters
        ----------
        filename : str, optional
            The filename to save the plot as (in PNG format). Default is None.
        show_properties : bool, optional
            Whether to show properties of the agents and modules in the plot. Default is True.
        dpi : int, optional
            The DPI (dots per inch) of the plot. Default is 300.
        truncate_cells : int, optional
            The number of cells to show in the plot. If there are more cells, the remaining cells will be truncated.
            Default is 10.

        Raises
        ------
        ImportError
            If Graphviz is not installed.
        """
        try:
            from graphviz import Digraph
            dot = Digraph()
            dot.attr(dpi=str(dpi))

            # Add environment; prepend "Base" to environment name if it is the base class
            env_name = type(self.environment).__name__
            env_label = f'Environment: {"Base" + env_name if env_name == "Environment" else ""}' + (
                        f'\nDimensions: {self.environment.dimensions}\nSize: {self.environment.size}'
                        if show_properties else '')
            dot.node('env', env_label, shape='rectangle')

            # Add organoid with superclass name
            organoid_label = f'Organoid: {type(self).__name__}' + (f'\nAgents: {len(self.agents)}'
                                                                   if show_properties else '')
            dot.node('organoid', organoid_label)
            dot.edge('env', 'organoid')

            # Add cells and their modules
            cell_ids = []
            do_truncate = (truncate_cells is not None) and (len(self.agents) > truncate_cells)
            for i, cell in enumerate(self.get_cells()):
                # If there are more than TRUNCATE_CELLS cells, show cells 1-[TRUNCATE-1], (...) node, and last cell
                if do_truncate and (i == truncate_cells-1):
                    ellipses_id = f'cell_{truncate_cells-1}'
                    dot.node(ellipses_id, '...', shape='none')
                    dot.edge('organoid', ellipses_id, style='invis')
                    cell_ids.append(ellipses_id)

                    ellipses_id = f'{truncate_cells-1}_mod_{0}'
                    dot.node(ellipses_id, '...', shape='none')
                    dot.edge(f'cell_{truncate_cells-1}', ellipses_id, style='invis')
                    dot.edge(ellipses_id, 'ml_model', style='invis')

                    cell = self.get_cells()[-1]
                    i = len(self.get_cells()) - 1

                cell_id = f'cell_{i}'
                cell_ids.append(cell_id)
                # Round each value in the position tuple to 2 decimal places
                cell_label = f'Cell {i+1}: {type(cell).__name__}' + (
                             f'\nPosition: {tuple(map(lambda x: round(x, 2), cell.position))}'
                             if show_properties else '')
                dot.node(cell_id, cell_label)
                dot.edge('organoid', cell_id)

                # Add modules to each cell
                for j, module in enumerate(cell.modules):
                    module_id = f'{cell_id}_mod_{j}'
                    module_label = f'Module {j+1}: {type(module).__name__}'
                    dot.node(module_id, module_label)
                    dot.edge(cell_id, module_id)

                    # Add ML model reference if available
                    if hasattr(module, 'ml_model'):
                        ml_model_input_shape = module.ml_model.input_shape
                        ml_model_id = 'ml_model'  # f'{module_id}_ml'
                        ml_model_label = type(module.ml_model).__name__ + (f'\nInput Shape: {ml_model_input_shape}'
                                                                           if show_properties else '')
                        dot.node(ml_model_id, ml_model_label)
                        with dot.subgraph() as s:
                            s.attr(rank='same')
                            s.node(ml_model_id, shape='rectangle')
                        dot.edge(module_id, ml_model_id, dir='back')

                if do_truncate and (i == len(self.get_cells()) - 1):
                    break

            with dot.subgraph() as s:
                s.attr(rank='same')
                for cell_id in cell_ids:
                    s.node(cell_id)

            if filename is not None:
                # If the filename has an extension, use it; otherwise, default to PNG
                ext = filename.split('.')[-1] if len(filename.split('.')) > 1 else 'png'
                filename = filename.split('.')[0]
                dot.render(filename, format=ext, cleanup=True)
                print(f'Organoid structure plot saved as "{filename}.{ext}"')
            else:
                dot.view()
        except ImportError as e:
            print("Error: Graphviz is required to plot the organoid structure.")
            raise e

    def plot_simulation_history(self, title, y_label, x_label="Time Steps", filename=None, dpi=300, figsize=(10, 6)):
        """
        Plot the simulation history of the organoid using Matplotlib.

        Parameters
        ----------
        title : str
            The title of the plot.
        y_label : str
            The label for the y-axis.
        x_label : str, optional
            The label for the x-axis. Default is "Time Steps".
        filename : str, optional
            The filename to save the plot as (in PNG format). Default is None (view the image with plt.show()).
        dpi : int, optional
            The DPI (dots per inch) of the plot. Default is 300.
        figsize : tuple, optional
            The size of the figure (width, height) in inches. Default is (10, 6).

        Raises
        ------
        ImportError
            If Matplotlib is not installed.
        """
        try:
            import numpy as np
            import matplotlib.pyplot as plt

            def contains_arrays(lst):
                """Check if the list contains any numpy arrays."""
                return any(isinstance(item, np.ndarray) for item in lst)

            def flatten_list(lst):
                """Flatten a list containing numpy arrays."""
                return [item if not isinstance(item, np.ndarray) else item.item() for item in lst]

            plt.figure(figsize=figsize)
            for i, cell in enumerate(self.get_cells()):
                history = cell.get_history()
                if contains_arrays(history):
                    history = flatten_list(history)
                plt.plot(history, label=f'Cell {i + 1}')

            plt.xlabel(x_label)
            plt.ylabel(y_label)
            plt.title(title)
            plt.legend()
            if filename is not None:
                plt.savefig(filename, dpi=dpi)
                print(f'Simulation history plot saved as "{filename}"')
            else:
                plt.show()
        except ImportError as e:
            print("Error: Matplotlib and NumPy are required to plot the simulation history.")
            raise e



import numpy as np
from .base import Cell


class SpikingNeuronCell(Cell):
    """
    A simple spiking neuron cell model (e.g., a Leaky Integrate-and-Fire neuron).
    The membrane potential is reset to zero after spiking.
    It is a subclass of the base Cell class.

    Parameters
    ----------
    position : {tuple, int}
        The position of the cell.
    threshold : float, optional
        The threshold for spiking. Default is 1.0.
    input_data_func : function, optional
        A function to generate input data for the neuron. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function
        The function to generate input data for the neuron.
    membrane_potential : float
        The membrane potential of the neuron.
    threshold : float
        The threshold for spiking.

    Methods
    -------
    update()
        Update the cell's state.
    get_history()
        Get the historical values of the cell.
    get_input_data()
        Get the input data for the neuron.
    spike(verbose=False)
        Spike the neuron.
    """
    def __init__(self, position, threshold=1.0, input_data_func=None):
        super().__init__(position, input_data_func=input_data_func or self.default_input_data_func)
        self.membrane_potential = 0.0
        self.threshold = threshold

    def spike(self, verbose=False):
        """
        Spike the neuron. The membrane potential is reset to zero after spiking.

        Parameters
        ----------
        verbose : bool, optional
            Whether to print verbose output. Default is False
        """
        self.membrane_potential = 0.0  # Reset potential after spiking
        if verbose:
            print(f"Neuron at {self.position} spiked!")

    def update(self):
        """
        Update the cell's state. Record the membrane potential and spike if it exceeds the threshold.
        """
        super().update(self.membrane_potential)  # Record the membrane potential
        if self.membrane_potential >= self.threshold:
            self.spike()

    @staticmethod
    def default_input_data_func():
        """
        Default input data function based on position.
        Simply returns a list of 10 input values (this should be overridden by providing an input data function).

        Returns
        -------
        list
            List of input data values.
        """
        return [0.5] * 10  # 10 input values


class GrowthShrinkageCell(Cell):
    """
    A simple cell model that grows and shrinks based on a volume parameter (e.g., in response to nutrient availability).
    It is a subclass of the base Cell class.

    Parameters
    ----------
    position : {tuple, int}
        The position of the cell.
    initial_volume : float, optional
        The initial volume of the cell. Default is 1.0.
    input_data_func : function, optional
        A function to generate input data for the cell. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function, optional
        A function to generate input data for the cell. Defaults to the get_volume method.
    volume : float
        The volume of the cell.

    Methods
    -------
    update()
        Update the cell's state.
    get_history()
        Get the historical values of the cell.
    get_input_data()
        Get the input data for the cell.
    grow(amount)
        Grow the cell by a specified amount.
    shrink(amount)
        Shrink the cell by a specified amount.
    get_volume()
        Get the volume of the cell.
    """
    def __init__(self, position, initial_volume=1.0, input_data_func=None):
        super().__init__(position, input_data_func=input_data_func or self.get_volume)
        self.volume = initial_volume

    def update(self):
        """
        Update the cell's state based on its volume.
        """
        super().update(self.volume)

    def grow(self, amount):
        """
        Grow the cell by a specified amount.

        Parameters
        ----------
        amount : float
            The amount to grow the cell by.
        """
        self.volume += amount

    def shrink(self, amount):
        """
        Shrink the cell by a specified amount.

        Parameters
        ----------
        amount : float
            The amount to shrink the cell by.
        """
        self.volume -= amount

    def get_volume(self, _=None):
        """
        Get the current volume of the cell.

        Returns
        -------
        float
            The volume of the cell.
        """
        return self.volume


class DifferentiatingCell(Cell):
    """
    A simple cell model that can differentiate into different states (e.g., an undifferentiated stem cell).
    It is a subclass of the base Cell class.

    Parameters
    ----------
    position : {tuple, int}
        The position of the cell.
    state : str, optional
        The initial state of the cell. Default is 'undifferentiated'.
    input_data_func : function, optional
        A function to generate input data for the cell. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function, optional
        A function to generate input data for the cell.
    state : str
        The state of the cell.

    Methods
    -------
    update()
        Update the cell's state.
    get_history()
        Get the historical values of the cell.
    get_input_data()
        Get the input data for the cell.
    differentiate(new_state)
        Differentiate the cell into a new state.
    get_state()
        Get the state of the cell.
    """
    def __init__(self, position, state='undifferentiated', input_data_func=None):
        super().__init__(position, input_data_func=input_data_func or self.get_state)
        self.state = state

    def update(self):
        """
        Update the cell's state based on its current state.
        """
        super().update(self.state)

    def differentiate(self, new_state):
        """
        Differentiate the cell into a new state.

        Parameters
        ----------
        new_state : str
            The new state of the cell.
        """
        self.state = new_state

    def get_state(self, _=None):
        """
        Get the current state of the cell.

        Returns
        -------
        str
            The state of the cell.
        """
        return self.state


class ChemotacticCell(Cell):
    """
    A simple cell model that can move towards a chemotactic gradient (e.g., in response to a chemical signal).
    It is a subclass of the base Cell class.

    Parameters
    ----------
    position : {tuple, int}
        The position of the cell.
    input_data_func : function, optional
        A function to generate input data for the cell. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function, optional
        A function to generate input data for the cell.
    chemotactic_gradient : float
        The chemotactic gradient for the cell.
        Positive values indicate movement towards the gradient.
        Negative values indicate movement away from the gradient.

    Methods
    -------
    update()
        Update the cell's state.
    get_history()
        Get the historical values of the cell.
    get_input_data()
        Get the input data for the cell.
    move_towards_gradient(gradient)
        Move the cell towards the specified gradient.
    get_chemotactic_gradient()
        Get the chemotactic gradient of the cell.
    """
    def __init__(self, position, input_data_func=None):
        super().__init__(position, input_data_func=input_data_func or self.get_chemotactic_gradient)
        self.chemotactic_gradient = 0.0

    def update(self):
        """
        Update the cell's state based on the chemotactic gradient.
        """
        super().update(self.chemotactic_gradient)

    def move_towards_gradient(self, gradient):
        """
        Move the cell towards the specified gradient.

        Parameters
        ----------
        gradient : float
            The gradient to move towards.
        """
        self.position += gradient

    def get_chemotactic_gradient(self, _=None):
        """
        Get the chemotactic gradient of the cell.

        Returns
        -------
        float
            The chemotactic gradient of the cell.
        """
        return self.chemotactic_gradient


class ImmuneCell(Cell):
    """
    A simple immune cell model that can be activated or deactivated (e.g., in response to an infection).
    It is a subclass of the base Cell class.

    Parameters
    ----------
    position : {tuple, int}
        The position of the cell.
    active : bool, optional
        Whether the cell is initially active. Default is False.
    input_data_func : function, optional
        A function to generate input data for the cell. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function, optional
        A function to generate input data for the cell.
    active : bool
        Whether the cell is currently active.

    Methods
    -------
    update()
        Update the cell's state.
    get_history()
        Get the historical values of the cell.
    get_input_data()
        Get the input data for the cell.
    activate()
        Activate the cell.
    deactivate()
        Deactivate the cell.
    is_active()
        Check if the cell is currently active or not.
    """
    def __init__(self, position, active=False, input_data_func=None):
        super().__init__(position, input_data_func=input_data_func or self.is_active)
        self.active = active

    def update(self):
        """
        Update the cell's state based on its activation status.
        """
        super().update(self.active)

    def activate(self):
        """
        Activate the cell.
        """
        self.active = True

    def deactivate(self):
        """
        Deactivate the cell.
        """
        self.active = False

    def is_active(self, _=None):
        """
        Check if the cell is currently active or not.

        Returns
        -------
        bool
            True if the cell is active, False otherwise.
        """
        return self.active


class SynapticPlasticityCell(Cell):
    """
    A simple cell model that can form synapses with other cells and transmit signals between them.
    The cells are connected by synapses that can change their strength based on the timing of spikes (e.g., STDP).
    It is a subclass of the base Cell class.

    Parameters
    ----------
    position : {tuple, int}
        The position of the cell.
    initial_synapses : list, optional
        List of initial synapses for the cell. Default is None.
    input_data_func : function, optional
        A function to generate input data for the cell. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function, optional
        A function to generate input data for the cell.
    synapses : list
        List of synapses connected to the cell.

    Methods
    -------
    update()
        Update the cell's state.
    get_history()
        Get the historical values of the cell.
    get_input_data()
        Get the input data for the cell.
    add_synapse(synapse)
        Add a synapse to the cell.
    get_synapses()
        Get the synapses connected to the cell.
    """
    def __init__(self, position, initial_synapses=None, input_data_func=None):
        super().__init__(position, input_data_func=input_data_func or self.get_synapses)
        self.synapses = initial_synapses if initial_synapses else []

    def update(self, historical_value=None):
        """
        Update the cell's state based on its synapses and historical value.
        Transmit signals between connected cells.

        Parameters
        ----------
        historical_value : any, optional
            The historical value to append to the cell's history. Default is None.
        """
        super().update(historical_value)
        for synapse in self.synapses:
            synapse.transmit()

    def add_synapse(self, synapse):
        """
        Add a synapse to the cell.
        """
        self.synapses.append(synapse)

    def get_synapses(self, _=None):
        """
        Get the synapses connected to the cell.

        Returns
        -------
        list
            List of synapses connected to the cell.
        """
        return self.synapses


class MetabolicCell(Cell):
    """
    A simple cell model that metabolizes energy based on a specified rate (e.g., in response to nutrient availability).
    It is a subclass of the base Cell class.

    Parameters
    ----------
    position : {tuple, int}
        The position of the cell.
    metabolism_rate : float, optional
        The rate at which the cell metabolizes energy. Default is 1.0.
    input_data_func : function, optional
        A function to generate input data for the cell. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function, optional
        A function to generate input data for the cell.
    metabolism_rate : float
        The rate at which the cell metabolizes energy.
    energy : float
        The energy level of the cell.

    Methods
    -------
    update()
        Update the cell's state.
    get_history()
        Get the historical values of the cell.
    get_input_data()
        Get the input data for the cell.
    metabolize(factor)
        Metabolize energy based on a specified factor.
    get_energy()
        Get the energy level of the cell.
    """
    def __init__(self, position, metabolism_rate=1.0, input_data_func=None):
        super().__init__(position, input_data_func=input_data_func or self.get_energy)
        self.metabolism_rate = metabolism_rate
        self.energy = 100.0

    def update(self):
        """
        Update the cell's state based on its energy level.
        """
        super().update(self.energy)

    def metabolize(self, factor):
        """
        Metabolize energy based on a specified factor.
        """
        self.energy *= factor

    def get_energy(self, _=None):
        """
        Get the energy level of the cell.

        Returns
        -------
        float
            The energy level of the cell.
        """
        return self.energy


class GeneRegulationCell(Cell):
    """
    A simple cell model that regulates gene expression levels based on external factors (e.g., signaling molecules).
    It is a subclass of the base Cell class.

    Parameters
    ----------
    position : {tuple, int}
        The position of the cell.
    gene_expression_level : float, optional
        The initial gene expression level. Default is 1.0.
    regulation_variance : float, optional
        The maximum variance in the regulation of gene expression level. Default is 0.05.
    input_data_func : function, optional
        A function to generate input data for the cell. Default is None.

    Attributes
    ----------
    position : {tuple, int}
        The position of the cell in the environment.
    modules : list
        List of modules that define the behavior of the cell.
    history : list
        List of historical values for the cell.
    input_data_func : function, optional
        A function to generate input data for the cell.
    gene_expression_level : float
        The gene expression level of the cell.
    regulation_variance : float
        The maximum variance in the regulation of gene expression level.

    Methods
    -------
    update()
        Update the cell's state.
    get_history()
        Get the historical values of the cell.
    get_input_data()
        Get the input data for the cell.
    regulate_genes(regulation_factor)
        Regulate the cell's gene expression level by a specified factor.
    get_gene_expression_level()
        Get the gene expression level of the cell.
    """
    def __init__(self, position, gene_expression_level=1.0, regulation_variance=0.05, input_data_func=None):
        super().__init__(position, input_data_func=input_data_func or self.get_gene_expression_level)
        self.gene_expression_level = gene_expression_level
        self.regulation_variance = regulation_variance

    def update(self):
        """
        Update the cell's state based on its gene expression level.
        """
        super().update(self.gene_expression_level)

    def regulate_genes(self, regulation_factor):
        """
        Regulate the cell's gene expression level by a specified factor.

        Parameters
        ----------
        regulation_factor : float
            The factor by which to regulate the gene expression level.
        """
        self.gene_expression_level *= regulation_factor
        self.gene_expression_level += np.random.uniform(-self.regulation_variance, self.regulation_variance)

    def get_gene_expression_level(self, _=None):
        """
        Get the gene expression level of the cell.

        Returns
        -------
        float
            The gene expression level of the cell.
        """
        return self.gene_expression_level



import numpy as np
from .cells import *
from .modules import *
from .base import Organoid, Synapse
from .utils import generate_random_position


class SpikingNeuronOrganoid(Organoid):
    def __init__(self, environment, ml_model, num_cells=10, input_data_func=None):
        """
    This class represents an organoid composed of spiking neurons. It is a subclass of the base Organoid class.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid resides.
    ml_model : BaseMLModel
        The machine learning model used by the spiking neuron modules.
    num_cells : int, optional
        The number of spiking neurons in the organoid. Default is 10.
    input_data_func : callable, optional
        A function that generates input data for the spiking neurons. Defaults to '() => [0.5] * 10' if None.
    """
        super().__init__(environment)
        for i in range(num_cells):
            neuron = SpikingNeuronCell(position=generate_random_position(environment.dimensions, environment.size),
                                       input_data_func=input_data_func)
            neuron.add_module(SpikingNeuronModule(ml_model))
            self.add_agent(neuron)


class GrowthShrinkageOrganoid(Organoid):
    """
    This class represents an organoid composed of cells that grow and shrink based on machine learning predictions.
    It is a subclass of the base Organoid class.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid resides.
    ml_model : BaseMLModel
        The machine learning model used by the growth/shrinkage cell modules.
    num_cells : int, optional
        The number of growth/shrinkage cells in the organoid. Default is 10.
    initial_cell_volume : float, optional
        The initial volume of the cells. Default is 1.0. Use "None" for random volumes in the range [0.5, 1.5].
    growth_amount : float, optional
        The amount by which the cells grow at each time step. Default is 0.1.
    growth_variance : float, optional
        The variance of the growth amount. Default is 0.05.
    prediction_threshold : float, optional
        The threshold above which the cells grow and below which they shrink. Default is 0.5.
    input_data_func : callable, optional
        A function that generates input data for the cell. Defaults to returning cell volume if None.
    """
    def __init__(self, environment, ml_model, num_cells=10, initial_cell_volume=1.0,
                 growth_amount=0.1, growth_variance=0.05, prediction_threshold=0.5, input_data_func=None):
        super().__init__(environment)
        for i in range(num_cells):
            initial_volume = initial_cell_volume if initial_cell_volume is not None else np.random.uniform(0.5, 1.5)
            cell = GrowthShrinkageCell(position=generate_random_position(environment.dimensions, environment.size),
                                       initial_volume=initial_volume, input_data_func=input_data_func)
            cell.add_module(GrowthShrinkageModule(ml_model, growth_amount, growth_variance, prediction_threshold))
            self.add_agent(cell)


class DifferentiationOrganoid(Organoid):
    """
    This class represents an organoid composed of differentiating (e.g., neural network) cells.
    It is a subclass of the base Organoid class.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid resides.
    ml_model : BaseMLModel
        The machine learning model used by the differentiating cell modules.
    num_cells : int, optional
        The number of differentiating cells in the organoid. Default is 10.
    input_data_func : callable, optional
        A function that generates input data for the cell. Defaults to returning cell state if None.
    """
    def __init__(self, environment, ml_model, num_cells=10, input_data_func=None):
        super().__init__(environment)
        for i in range(num_cells):
            cell = DifferentiatingCell(position=generate_random_position(environment.dimensions, environment.size),
                                       input_data_func=input_data_func)
            cell.add_module(DifferentiationModule(ml_model))
            self.add_agent(cell)


class ChemotaxisOrganoid(Organoid):
    """
    This class represents an organoid composed of chemotactic cells. It is a subclass of the base Organoid class.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid resides.
    ml_model : BaseMLModel
        The machine learning model used by the chemotactic cell modules.
    num_cells : int, optional
        The number of chemotactic cells in the organoid. Default is 10.
    input_data_func : callable, optional
        A function that generates input data for the cell. Defaults to returning cell gradient if None.
    """
    def __init__(self, environment, ml_model, num_cells=10, input_data_func=None):
        super().__init__(environment)
        for i in range(num_cells):
            cell = ChemotacticCell(position=generate_random_position(environment.dimensions, environment.size),
                                   input_data_func=input_data_func)
            cell.add_module(ChemotaxisModule(ml_model))
            self.add_agent(cell)


class ImmuneResponseOrganoid(Organoid):
    """
    This class represents an organoid composed of immune cells that respond to external stimuli.
    It is a subclass of the base Organoid class.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid resides.
    ml_model : BaseMLModel
        The machine learning model used by the immune cell modules.
    num_cells : int, optional
        The number of immune cells in the organoid. Default is 10.
    input_data_func : callable, optional
        A function that generates input data for the cell. Defaults to returning cell active state if None.
    """
    def __init__(self, environment, ml_model, num_cells=10, input_data_func=None):
        super().__init__(environment)
        for i in range(num_cells):
            cell = ImmuneCell(position=generate_random_position(environment.dimensions, environment.size),
                              input_data_func=input_data_func)
            cell.add_module(ImmuneResponseModule(ml_model))
            self.add_agent(cell)


class SynapticPlasticityOrganoid(Organoid):
    """
    This class represents an organoid composed of neurons with synaptic plasticity (STDP).
    It is a subclass of the base Organoid class.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid resides.
    ml_model : BaseMLModel
        The machine learning model used by the synaptic plasticity cell modules.
    num_cells : int, optional
        The number of synaptic plasticity cells (i.e., neurons) in the organoid. Default is 10.
    num_synapses : int, optional
        The number of synapses connecting the plasticity cells. Default is 5.
    input_data_func : callable, optional
        A function that generates input data for the cell. Defaults to returning synapses if None.
    """
    def __init__(self, environment, ml_model, num_cells=10, num_synapses=5, input_data_func=None):
        super().__init__(environment)
        for i in range(num_cells):
            cell = SynapticPlasticityCell(position=generate_random_position(environment.dimensions, environment.size),
                                          input_data_func=input_data_func)
            cell.add_module(SynapticPlasticityModule(ml_model, synapse=None))
            self.add_agent(cell)

        for _ in range(num_synapses):
            pre_cell, post_cell = np.random.choice(self.agents, 2, replace=False)
            synapse = Synapse(pre_neuron=pre_cell, post_neuron=post_cell)
            pre_cell.add_synapse(synapse)
            post_cell.add_module(SynapticPlasticityModule(ml_model, synapse=synapse))


class MetabolicOrganoid(Organoid):
    """
    This class represents an organoid composed of cells with metabolic modules.
    It is a subclass of the base Organoid class.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid resides.
    ml_model : BaseMLModel
        The machine learning model used by the metabolic cell modules.
    num_cells : int, optional
        The number of metabolic cells in the organoid. Default is 10.
    input_data_func : callable, optional
        A function that generates input data for the cell. Defaults to returning cell energy if None.
    """
    def __init__(self, environment, ml_model, num_cells=10, input_data_func=None):
        super().__init__(environment)
        for i in range(num_cells):
            cell = MetabolicCell(position=generate_random_position(environment.dimensions, environment.size),
                                 input_data_func=input_data_func)
            cell.add_module(MetabolicModule(ml_model))
            self.add_agent(cell)


class GeneRegulationOrganoid(Organoid):
    """
    This class represents an organoid composed of cells with gene regulation modules.
    It is a subclass of the base Organoid class.
    NOTE: In the future, this class could be extended to include sequence data and gene expression profiles.

    Parameters
    ----------
    environment : Environment
        The environment in which the organoid resides.
    ml_model : BaseMLModel
        The machine learning model used by the gene regulation cell modules.
    num_cells : int, optional
        The number of gene regulation cells in the organoid. Default is 10.
    regulation_variance : float, optional
        The maximum variance of the gene regulation. Default is 0.05.
    input_data_func : callable, optional
        A function that generates input data for the cell. Defaults to returning cell gene expression if None.
    """
    def __init__(self, environment, ml_model, num_cells=10, regulation_variance=0.05, input_data_func=None):
        super().__init__(environment)
        for i in range(num_cells):
            cell = GeneRegulationCell(position=generate_random_position(environment.dimensions, environment.size),
                                      regulation_variance=regulation_variance, input_data_func=input_data_func)
            cell.add_module(GeneRegulationModule(ml_model))
            self.add_agent(cell)



try:
    import torch
    import torch.nn as nn
    from .models import BaseMLModel
    from .modules import BaseMLModule


    class TorchModel(BaseMLModel):
        """
        This adapter class represents a machine learning model implemented in PyTorch.
        It is a subclass of the BaseMLModel class.
        The PyTorch addon for pyorganoid must be installed using the pyorganoid[all] package or the command:
            pip install pyorganoid[torch]

        Parameters
        ----------
        model_path : str
            The path to the PyTorch model to use for prediction.
        model_class : type[torch.nn.Module]
            The PyTorch model class to use for prediction.
        input_shape : tuple, optional
            The input shape of the model. Default is None (infer from model if possible).

        Attributes
        ----------
        model : torch.nn.Module
            The PyTorch model to use for prediction.
        input_shape : tuple
            The input shape of the model.

        Methods
        -------
        predict(input_data, verbose=False)
            Predict the output of the model given an input. Returns the predicted output.

        Raises
        ------
        ValueError
            If the model_class is not provided or is None.
        """
        def __init__(self, model_path, model_class, input_shape=None):
            if model_class is None:
                raise ValueError("You must provide a model_class to load the model.")
            self.input_shape = input_shape
            if self.input_shape is None:
                self.input_shape = self.__get_input_shape()
            self.model = model_class(input_shape[0])
            self.model.load_state_dict(torch.load(model_path))
            self.model.eval()

        def predict(self, input_data, verbose=False):
            """
            Predict the output of the model given an input.

            Parameters
            ----------
            input_data : array-like
                The input data for the model.
            verbose : bool, optional
                Whether to print verbose output. Default is False.

            Returns
            -------
            array-like
                The output of the model given the input data.
            """
            with torch.no_grad():
                # input_tensor = torch.tensor(input_data).float().unsqueeze(0)
                # Reshape input data to match model input shape as necessary (if possible)
                input_tensor = torch.tensor(input_data).float()
                if self.input_shape is not None:
                    input_tensor = input_tensor.view(1, *self.input_shape)
                output = self.model(input_tensor)
                return output.numpy()

        def __get_input_shape(self):
            """
            Attempt to determine the input shape of the model.

            Returns
            -------
            tuple
                The input shape of the model.

            Raises
            ------
            ValueError
                If the input shape cannot be determined from the model.
                If this occurs, you must provide the input_shape manually when creating the TorchModel instance.
            """
            if hasattr(self.model, 'input_shape'):
                return self.model.input_shape
            else:
                # Attempt to determine the input shape from the first layer
                for module in self.model.modules():
                    if isinstance(module, (nn.Linear, nn.Conv2d)):
                        return module.in_features if isinstance(module, nn.Linear) else module.in_channels,
                raise ValueError("Unable to determine input shape from the model. Please provide input_shape manually.")


    class TorchModule(BaseMLModule):
        pass
except ImportError:
    TFModel = None
    TFModule = None



import numpy as np


class BaseModule:
    """
    Base class for all modules (not to be instantiated directly).
    Modules are used to implement the logic of the agents in the simulation.
    Subclasses should implement the run method.

    Methods
    -------
    run(agent)
        Run the module on the given agent.
    """
    def __init__(self, *args, **kwargs):
        pass

    def run(self, agent):
        """
        Run the module on the given agent.

        Parameters
        ----------
        agent : Agent
            The agent to run the module on.
        """
        raise NotImplementedError("Subclasses should implement this method!")


class BaseMLModule(BaseModule):
    """
    Base class for machine learning modules (not to be instantiated directly).
    Subclasses should implement the collect_input_data (optionally) and apply_prediction methods.
    It is a subclass of the BaseModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Attributes
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Methods
    -------
    run(agent, verbose=False)
        Run the module on the given agent.
    collect_input_data(agent)
        Collect the input data for the machine learning model.
    apply_prediction(agent, prediction)
        Apply the prediction of the machine learning model to the agent.
    """
    def __init__(self, ml_model):
        super().__init__()
        self.ml_model = ml_model

    def run(self, agent, verbose=False):
        """
        Run the module on the given agent. Collects input data, makes a prediction, and applies it to the agent.

        Parameters
        ----------
        agent : Agent
            The agent to run the module on.
        verbose : bool, optional
            Whether to print verbose output. Default is False.
        """
        input_data = self.collect_input_data(agent)
        prediction = self.ml_model.predict(input_data, verbose=verbose)
        self.apply_prediction(agent, prediction)

    @staticmethod
    def collect_input_data(agent):
        """
        Collect the input data for the machine learning model.

        Parameters
        ----------
        agent : Agent
            The agent to collect input data from.

        Returns
        -------
        array-like
            The input data for the machine learning model.
        """
        return [agent.get_input_data(agent)]  # Default implementation, can be overridden

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.

        Raises
        ------
        NotImplementedError
            If the subclass does not implement this method.
        """
        raise NotImplementedError("Subclasses should implement this method!")

    def drop_prediction_dimensionality(self, prediction):
        """
        Drop the dimensionality of the prediction if necessary (e.g., from Tensorflow models).

        Parameters
        ----------
        prediction : array-like
            The prediction output of the machine learning model.

        Returns
        -------
        array-like
            The prediction output with reduced dimensionality.
        """
        try:
            from .tf_module import TFModel
            from .onnx_module import ONNXModel
            from .torch_module import TorchModel
            if isinstance(self.ml_model, TFModel):
                return prediction[0] if isinstance(prediction, list) and len(prediction) > 0 else prediction
            elif isinstance(self.ml_model, TorchModel):
                return prediction[0] if ((isinstance(prediction, list) or isinstance(prediction, np.ndarray))
                                         and len(prediction) > 0) else prediction
            elif isinstance(self.ml_model, ONNXModel):
                if isinstance(prediction, list) and len(prediction) > 0:
                    prediction = prediction[0]
                if isinstance(prediction, np.ndarray):
                    prediction = prediction.flatten().tolist()
                if isinstance(prediction, float) or isinstance(prediction, int):
                    return [prediction]
                return prediction
            else:
                return prediction
        except ImportError:
            return prediction


class SpikingNeuronModule(BaseMLModule):
    """
    Module for simulating a spiking neuron using a machine learning model.
    The module collects the membrane potential of the agent as input and
    applies the prediction to the membrane potential.
    It is a subclass of the BaseMLModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Attributes
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Methods
    -------
    run(agent, verbose=False)
        Run the module on the given agent.
    collect_input_data(agent)
        Collect the input data for the machine learning model.
    apply_prediction(agent, prediction)
        Apply the prediction of the machine learning model to the agent.
    """

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.
        Adds the prediction to the membrane potential of the agent (neuron).

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.
        """
        prediction = self.drop_prediction_dimensionality(prediction)
        agent.membrane_potential += prediction[0]


class GrowthShrinkageModule(BaseMLModule):
    """
    Module for simulating growth and shrinkage of cells using a machine learning model.
    The module collects the volume of the agent as input and applies the prediction to the volume.
    It is a subclass of the BaseMLModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.
    growth_amount : float, optional
        The amount by which the cells grow at each time step. Default is 0.1.
    growth_variance : float, optional
        The variance of the growth amount. Default is 0.05.
    prediction_threshold : float, optional
        The threshold above which the cells grow and below which they shrink. Default is 0.5.

    Attributes
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.
    growth_amount : float, optional
        The amount by which the cells grow at each time step.
    growth_variance : float, optional
        The variance of the growth amount.
    prediction_threshold : float, optional
        The threshold above which the cells grow and below which they shrink.

    Methods
    -------
    run(agent, verbose=False)
        Run the module on the given agent.
    collect_input_data(agent)
        Collect the input data for the machine learning model.
    apply_prediction(agent, prediction)
        Apply the prediction of the machine learning model to the agent.
    """
    def __init__(self, ml_model, growth_amount=0.1, growth_variance=0.05, prediction_threshold=0.5):
        super().__init__(ml_model)
        self.growth_amount = growth_amount
        self.growth_variance = growth_variance
        self.prediction_threshold = prediction_threshold

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.
        If the prediction is greater than 0.5, the agent (cell) grows; otherwise, it shrinks.

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.
        """
        prediction = self.drop_prediction_dimensionality(prediction)
        change = self.growth_amount + np.random.uniform(-self.growth_variance, self.growth_variance)
        if prediction[0] > 0.5:
            agent.grow(change)
        else:
            agent.shrink(change)


class DifferentiationModule(BaseMLModule):
    """
    Module for simulating differentiation of cells (such as in a Neural Network) using a machine learning model.
    The module collects the state of the agent as input and applies the prediction to the state.
    It is a subclass of the BaseMLModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Attributes
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Methods
    -------
    run(agent, verbose=False)
        Run the module on the given agent.
    collect_input_data(agent)
        Collect the input data for the machine learning model.
    apply_prediction(agent, prediction)
        Apply the prediction of the machine learning model to the agent.
    """

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.
        Differentiates the agent (cell) based on the prediction.

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.
        """
        prediction = self.drop_prediction_dimensionality(prediction)
        agent.differentiate(prediction[0])


class ChemotaxisModule(BaseMLModule):
    """
    Module for simulating chemotaxis of cells using a machine learning model.
    The module collects the chemotactic gradient of the agent as input and applies the prediction to the gradient.
    It is a subclass of the BaseMLModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Attributes
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Methods
    -------
    run(agent, verbose=False)
        Run the module on the given agent.
    collect_input_data(agent)
        Collect the input data for the machine learning model.
    apply_prediction(agent, prediction)
        Apply the prediction of the machine learning model to the agent.
    """

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.
        Moves the agent (cell) towards the predicted gradient.

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.
        """
        prediction = self.drop_prediction_dimensionality(prediction)
        agent.move_towards_gradient(prediction[0])


class ImmuneResponseModule(BaseMLModule):
    """
    Module for simulating an immune response using a machine learning model.
    The module collects whether the agent is active or not as input and applies the prediction to the agent.
    It is a subclass of the BaseMLModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Attributes
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Methods
    -------
    run(agent, verbose=False)
        Run the module on the given agent.
    collect_input_data(agent)
        Collect the input data for the machine learning model.
    apply_prediction(agent, prediction)
        Apply the prediction of the machine learning model to the agent.
    """

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.
        Activates or deactivates the agent based on the prediction.

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.
        """
        prediction = self.drop_prediction_dimensionality(prediction)
        if prediction[0] > 0.5:
            agent.activate()
        else:
            agent.deactivate()


class SynapticPlasticityModule(BaseMLModule):
    """
    Module for simulating spike-timing-dependent plasticity (STDP) in a synapse (e.g., in a physical neural network).
    The module updates the synaptic weight based on the pre- and post-synaptic spikes.
    It is a subclass of the BaseMLModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.
    synapse : Synapse
        The synapse to update.
    learning_rate : float, optional
        The learning rate for updating the synaptic weight. Default is 0.01.

    Attributes
    ----------
    synapse : Synapse
        The synapse to update.
    learning_rate : float
        The learning rate for updating the synaptic weight.

    Methods
    -------
    run(agent)
        Run the module on the given agent.

    """
    def __init__(self, ml_model, synapse, learning_rate=0.01):
        super().__init__(ml_model)
        self.synapse = synapse
        self.learning_rate = learning_rate

    def run(self, agent):
        pre_spike = self.synapse.pre_neuron.membrane_potential >= self.synapse.pre_neuron.threshold
        post_spike = self.synapse.post_neuron.membrane_potential >= self.synapse.post_neuron.threshold

        if pre_spike and post_spike:
            self.synapse.weight += self.learning_rate * (1 - self.synapse.weight)
        elif pre_spike and not post_spike:
            self.synapse.weight -= self.learning_rate * self.synapse.weight

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.
        Adds the prediction to the membrane potential of the agent (neuron).

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.
        """
        prediction = self.drop_prediction_dimensionality(prediction)
        agent.membrane_potential += prediction[0]


class MetabolicModule(BaseMLModule):
    """
    Module for simulating metabolism in cells using a machine learning model.
    The module collects the energy level of the agent as input and applies the prediction to the agent.
    It is a subclass of the BaseMLModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Attributes
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Methods
    -------
    run(agent, verbose=False)
        Run the module on the given agent.
    collect_input_data(agent)
        Collect the input data for the machine learning model.
    apply_prediction(agent, prediction)
        Apply the prediction of the machine learning model to the agent.
    """

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.
        Metabolizes the agent based on the prediction.

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.
        """
        metabolism_factor = prediction[0]
        agent.metabolize(metabolism_factor)


class GeneRegulationModule(BaseMLModule):
    """
    Module for simulating gene regulation in cells using a machine learning model.
    The module collects the gene expression level of the agent as input and applies the prediction to the agent.
    It is a subclass of the BaseMLModule class.

    Parameters
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Attributes
    ----------
    ml_model : BaseMLModel
        The machine learning model to use for prediction.

    Methods
    -------
    run(agent, verbose=False)
        Run the module on the given agent.
    collect_input_data(agent)
        Collect the input data for the machine learning model.
    apply_prediction(agent, prediction)
        Apply the prediction of the machine learning model to the agent.
    """

    def apply_prediction(self, agent, prediction):
        """
        Apply the prediction of the machine learning model to the agent.
        Regulates the genes of the agent based on the prediction.

        Parameters
        ----------
        agent : Agent
            The agent to apply the prediction to.
        prediction : array-like
            The prediction output of the machine learning model.
        """
        prediction = self.drop_prediction_dimensionality(prediction)
        regulation_factor = prediction[0]
        scaled_factor = 0.5 + regulation_factor  # Scale regulation_factor from [0, 1] to [0.5, 1.5]
        print(f"Regulation factor: {scaled_factor}")
        agent.regulate_genes(scaled_factor)



class BaseMLModel:
    """
    This is the base adapter class for machine learning models used in pyorganoid (not to be instantiated directly).
    Subclasses should implement the predict method.

    Methods
    -------
    predict(input_data, verbose=False)
        Predict the output of the model given an input.
    """
    def predict(self, input_data, verbose=False):
        """
        Predict the output of the model given an input.

        Parameters
        ----------
        input_data : array-like
            The input data for the model.
        verbose : bool, optional
            Whether to print verbose output. Default is False.

        Returns
        -------
        array-like
            The output of the model given the input data.

        Raises
        ------
        NotImplementedError
            If the subclass does not implement this method.
        """
        raise NotImplementedError("Subclasses should implement this method!")



class AgentContainer:
    """
    Container for agents in the simulation.

    Attributes
    ----------
    agents : list
        List of agents in the simulation.

    Methods
    -------
    add_agent(agent)
        Add an agent to the container.
    get_agents()
        Get all agents in the container.
    get_agent_by_id(agent_id)
        Get an agent by its ID.
    """
    def __init__(self):
        self.agents = []

    def add_agent(self, agent):
        """
        Add an agent to the container.

        Parameters
        ----------
        agent : Agent
            The agent to add to the container.
        """
        self.agents.append(agent)

    def get_agents(self):
        """
        Get all agents in the container.

        Returns
        -------
        list
            List of agents in the container.
        """
        return self.agents

    def get_agent_by_id(self, agent_id):
        """
        Get an agent from the container by its ID.

        Parameters
        ----------
        agent_id : int
            The ID of the agent to get.

        Returns
        -------
        Agent
            The agent with the given ID, or None if not found.
        """
        return next((agent for agent in self.agents if agent.id == agent_id), None)


class AgentHandle:
    """
    Handle for an agent in the simulation.

    Attributes
    ----------
    agent_id : int
        The ID of the agent.
    container : AgentContainer
        The container that holds the agent.

    Methods
    -------
    get_agent()
        Get the agent from the container.
    """
    def __init__(self, agent_id, container):
        self.agent_id = agent_id
        self.container = container

    def get_agent(self):
        """
        Get the agent from the container.

        Returns
        -------
        Agent
            The agent with the given ID.
        """
        return self.container.get_agent_by_id(self.agent_id)



import numpy as np


class Scheduler:
    """
    Base class for simulation schedulers. Can be used to simulate the behavior of an organoid in a given environment.

    Parameters
    ----------
    organoid : Organoid
        The organoid to simulate.

    Attributes
    ----------
    organoid : Organoid
        The organoid to simulate.

    Methods
    -------
    simulate(steps)
        Simulate the organoid's behavior over a number of steps.
    """
    def __init__(self, organoid):
        self.organoid = organoid

    def simulate(self, steps):
        """
        Simulate the organoid's behavior over a number of steps.

        Parameters
        ----------
        steps : int
            The number of steps to run the simulation.
        """
        for step in range(steps):
            print(f"Step {step+1}/{steps}")
            self.organoid.environment.update()
            for agent in self.organoid.agents:
                agent.update()


class StochasticScheduler(Scheduler):
    """
    Scheduler that simulates the organoid's behavior with a stochastic update rule.
    The update rule is applied to each agent with a 50% chance.
    It is a subclass of the base Scheduler class.

    Parameters
    ----------
    organoid : Organoid
        The organoid to simulate.

    Attributes
    ----------
    organoid : Organoid
        The organoid to simulate.

    Methods
    -------
    simulate(steps)
        Simulate the organoid's behavior over a number of steps with a stochastic update rule.
    """
    def simulate(self, steps):
        """
        Simulate the organoid's behavior over a number of steps with a stochastic update rule.
        The update rule is applied to each agent with a 50% chance.

        Parameters
        ----------
        steps : int
            The number of steps to run the simulation.
        """
        for step in range(steps):
            print(f"Step {step+1}/{steps}")
            self.organoid.environment.update()
            for agent in self.organoid.agents:
                if np.random.rand() > 0.5:  # 50% chance to update the agent
                    agent.update()


class PriorityScheduler(Scheduler):
    """
    Scheduler that simulates the organoid's behavior based on a priority system.
    Agents are updated in order of decreasing priority.
    It is a subclass of the base Scheduler class.

    Parameters
    ----------
    organoid : Organoid
        The organoid to simulate.
    priorities : dict
        A dictionary of agent priorities.

    Attributes
    ----------
    organoid : Organoid
        The organoid to simulate.
    priorities : dict
        A dictionary of agent priorities.

    Methods
    -------
    simulate(steps)
        Simulate the organoid's behavior over a number of steps based on a priority system.
    """
    def __init__(self, organoid, priorities):
        super().__init__(organoid)
        self.priorities = priorities

    def simulate(self, steps):
        """
        Simulate the organoid's behavior over a number of steps based on a priority system.
        Agents are updated in order of decreasing priority.

        Parameters
        ----------
        steps : int
            The number of steps to run the simulation.
        """
        for step in range(steps):
            print(f"Step {step+1}/{steps}")
            self.organoid.environment.update()
            sorted_agents = sorted(self.organoid.agents, key=lambda agent: self.priorities.get(agent, 0), reverse=True)
            for agent in sorted_agents:
                agent.update()


class ParallelScheduler(Scheduler):
    """
    Scheduler that runs the simulation in parallel using multiple CPU cores.
    The class has not yet been implemented.
    It is a subclass of the base Scheduler class.

    Parameters
    ----------
    organoid : Organoid
        The organoid to simulate.

    Attributes
    ----------
    organoid : Organoid
        The organoid to simulate.

    Methods
    -------
    simulate(steps)
        Simulate the organoid's behavior over a number of steps.
    """

    def simulate(self, steps):
        """
        Simulate the organoid's behavior over a number of steps.

        This method is intended to run the simulation in parallel using multiple
        CPU cores. Currently, this function is not yet implemented. Using joblib may
        be the most feasible way to implement parallel simulation over multiprocessing;
        however, the cell history is neither preserved nor updated correctly when executed in parallel.

        Parameters
        ----------
        steps : int
            The number of steps to run the simulation.

        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError("Parallel simulation is not yet implemented. Please use a different scheduler.")



try:
    import joblib
    import sklearn
    import numpy as np
    from .models import BaseMLModel
    from .modules import BaseMLModule


    class SklearnModel(BaseMLModel):
        """
        This adapter class represents a machine learning model implemented in scikit-learn.
        It is a subclass of the BaseMLModel class.
        The scikit-learn addon for pyorganoid must be installed using the pyorganoid[all] package or the command:
            pip install pyorganoid[sklearn]

        Parameters
        ----------
        model_path : str
            The path to the scikit-learn model to use for prediction.
        num_features : int, optional
            The number of features expected by the model. Default is None (infer from model if possible).

        Attributes
        ----------
        model : sklearn.base.BaseEstimator
            The scikit-learn model to use for prediction.
        input_shape : tuple
            The number of features expected by the model.

        Methods
        -------
        predict(input_data, verbose=False)
            Predict the output of the model given an input. Returns the predicted output.
        """

        def __init__(self, model_path, num_features=None):
            self.model = joblib.load(model_path)
            self.input_shape = (num_features,)
            if self.input_shape is None and hasattr(self.model, 'n_features_in_'):
                self.input_shape = (self.model.n_features_in_,)

        def predict(self, input_data, verbose=False):
            """
            Predict the output of the model given an input.

            Parameters
            ----------
            input_data : array-like
                The input data for the model.
            verbose : bool, optional
                Whether to print verbose output. Default is False.

            Returns
            -------
            array-like
                The output of the model given the input data.
            """
            input_data = np.array(input_data).reshape((1, -1))
            return self.model.predict(input_data)


    class SklearnModule(BaseMLModule):
        pass
except ImportError:
    TFModel = None
    TFModule = None



import numpy as np


def generate_random_position(dimensions, size):
    """
    Generate a random position in the given dimensions and size.

    Parameters
    ----------
    dimensions : int
        The number of dimensions.
    size : float
        The size of the space.

    Returns
    -------
    np.ndarray
        A random position in the given dimensions
    """
    return np.random.rand(dimensions) * size



try:
    import numpy as np
    import tensorflow as tf
    from .models import BaseMLModel
    from .modules import BaseMLModule


    class TFModel(BaseMLModel):
        """
        This adapter class represents a machine learning model implemented in Tensorflow/Keras.
        It is a subclass of the BaseMLModel class.
        The Tensorflow addon for pyorganoid must be installed using the pyorganoid[all] package or the command:
            pip install pyorganoid[tensorflow]

        Parameters
        ----------
        model_path : str
            The path to the Tensorflow/Keras model to use for prediction.
        input_shape : tuple, optional
            The input shape of the model. Default is None (infer from model if possible).

        Attributes
        ----------
        model : tf.keras.Model
            The Tensorflow/Keras model to use for prediction.
        input_shape : tuple
            The input shape of the model.

        Methods
        -------
        predict(input_data, verbose=False)
            Predict the output of the model given an input. Returns the predicted output.
        """
        def __init__(self, model_path, input_shape=None):
            self.model = tf.keras.models.load_model(model_path)
            self.input_shape = input_shape
            if self.input_shape is None and self.model.input_shape is not None and len(self.model.input_shape) > 0:
                self.input_shape = self.model.input_shape[1:]

        def predict(self, input_data, verbose=False):
            """
            Predict the output of the model given an input.

            Parameters
            ----------
            input_data : array-like
                The input data for the model.
            verbose : bool, optional
                Whether to print verbose output. Default is False.

            Returns
            -------
            array-like
                The output of the model given the input data.
            """
            # Reshape input data to match model input shape as necessary (if possible)
            input_data = np.array(input_data)
            if self.input_shape is not None and input_data.shape != self.input_shape:
                input_data = input_data.reshape((1, *self.input_shape))  # .reshape((1, -1))
            return self.model.predict(input_data, verbose=verbose)


    class TFModule(BaseMLModule):
        pass
except ImportError:
    TFModel = None
    TFModule = None



from setuptools import setup, find_packages

# Be sure to update the version number in both "setup.py" and "meta.yaml" files.
setup(
    name='pyorganoid',
    version='0.1.2',
    packages=find_packages(),
    author='Daniel Szelogowski',
    description='A Python package for the simulation of organoids for the purpose of studying '
                'Organoid Intelligence (OI) and Organoid Learning (OL).',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://danielathome19.github.io/pyorganoid',
    project_urls={
        'Documentation': 'https://danielathome19.github.io/pyorganoid',
        'Source': 'https://github.com/danielathome19/pyorganoid',
        'Tracker': 'https://github.com/danielathome19/pyorganoid/issues',
    },
    install_requires=[
        'numpy',
        'matplotlib',
        'graphviz',
    ],
    extras_require={
        "tensorflow": ["tensorflow<=2.15"],
        "torch": ["torch"],
        "sklearn": ["scikit-learn", "joblib"],
        "onnx": ["onnxruntime", "onnxmltools"],
        "all": ["tensorflow<=2.15", "torch", "scikit-learn", "joblib", "onnxruntime", "onnxmltools"]
    },
    entry_points={
        'console_scripts': [
            'pyorganoid=pyorganoid:main',
        ],
    },
)



try:
    import numpy as np
    import onnxruntime as ort
    from .models import BaseMLModel
    from .modules import BaseMLModule


    class ONNXModel(BaseMLModel):
        """
        This adapter class represents a machine learning model implemented saved as an ONNX model.
        It is a subclass of the BaseMLModel class.
        The Tensorflow addon for pyorganoid must be installed using the pyorganoid[all] package or the command:
            pip install pyorganoid[onnx]

        Parameters
        ----------
        model_path : str
            The path to the ONNX model to use for prediction.
        input_shape : tuple, optional
            The input shape of the model. Default is None (infer from model if possible).
        data_type : str, optional
            The data type to use for input data. Default is 'float32'.

        Attributes
        ----------
        model : onnxruntime.InferenceSession
            The ONNX model to use for prediction.
        input_shape : tuple
            The input shape of the model.
        data_type : str
            The data type to use for input data.

        Methods
        -------
        predict(input_data, verbose=False)
            Predict the output of the model given an input. Returns the predicted output.
        """
        def __init__(self, model_path, input_shape=None, data_type='float32'):
            self.model = ort.InferenceSession(model_path)
            self.input_shape = input_shape
            if self.input_shape is None:
                self.input_shape = self.model.get_inputs()[0].shape
            self.data_type = data_type

        def predict(self, input_data, verbose=False):
            """
            Predict the output of the model given an input.

            Parameters
            ----------
            input_data : array-like
                The input data for the model.
            verbose : bool, optional
                Whether to print verbose output. Default is False.

            Returns
            -------
            array-like
                The output of the model given the input data.

            Raises
            ------
            ValueError
                If the input data shape does not match the model input shape (e.g., if the input_shape is 'None').
            """
            input_data = np.array(input_data, dtype=self.data_type)
            # Reshape input data to match model input shape as necessary (if possible)
            if self.input_shape is None:
                raise ValueError("Input shape must be specified for ONNX model")
            if self.input_shape is not None:
                input_data = input_data.reshape((1, *self.input_shape))
            return self.model.run(None, {self.model.get_inputs()[0].name: input_data})


    class ONNXModule(BaseMLModule):
        pass
except ImportError:
    ONNXModel = None
    ONNXModule = None



from .base import Agent, Cell, Synapse, Organoid
from .models import BaseMLModel
from .cells import (
    SpikingNeuronCell, GrowthShrinkageCell, DifferentiatingCell, ChemotacticCell,
    ImmuneCell, SynapticPlasticityCell, MetabolicCell, GeneRegulationCell
)
from .modules import (
    BaseModule, BaseMLModule, SpikingNeuronModule, GrowthShrinkageModule, DifferentiationModule, ChemotaxisModule,
    ImmuneResponseModule, SynapticPlasticityModule, MetabolicModule, GeneRegulationModule
)
from .organoids import (
    SpikingNeuronOrganoid, GrowthShrinkageOrganoid, DifferentiationOrganoid, ChemotaxisOrganoid,
    ImmuneResponseOrganoid, SynapticPlasticityOrganoid, MetabolicOrganoid, GeneRegulationOrganoid
)
from .environments import (
    Environment, GradientEnvironment, TemperatureEnvironment, StochasticEnvironment,
    ChemicalGradientEnvironment, ElectricFieldEnvironment
)
from .simulation import Scheduler, StochasticScheduler, PriorityScheduler, ParallelScheduler
from .agents import AgentContainer, AgentHandle
from .utils import generate_random_position
# TODO: bioinformatics and cheminformatics modules/importing, e.g. sequence alignment, molecular docking, etc.;
#       parallel/distributed scheduling; logging; visualization; etc.


__all__ = [
    'Agent', 'Cell', 'Synapse', 'Organoid',
    'SpikingNeuronCell', 'GrowthShrinkageCell', 'DifferentiatingCell', 'ChemotacticCell', 'ImmuneCell', 'SynapticPlasticityCell', 'MetabolicCell', 'GeneRegulationCell',
    'SpikingNeuronOrganoid', 'GrowthShrinkageOrganoid', 'DifferentiationOrganoid', 'ChemotaxisOrganoid', 'ImmuneResponseOrganoid', 'SynapticPlasticityOrganoid', 'MetabolicOrganoid', 'GeneRegulationOrganoid',
    'BaseModule', 'SpikingNeuronModule', 'GrowthShrinkageModule', 'DifferentiationModule', 'ChemotaxisModule', 'ImmuneResponseModule', 'SynapticPlasticityModule', 'MetabolicModule', 'GeneRegulationModule',
    'BaseMLModel',
    'Environment', 'GradientEnvironment', 'TemperatureEnvironment', 'StochasticEnvironment', 'ChemicalGradientEnvironment', 'ElectricFieldEnvironment',
    'Scheduler', 'StochasticScheduler', 'PriorityScheduler', 'ParallelScheduler',
    'AgentContainer', 'AgentHandle',
    'generate_random_position',
    # Conditional inclusion based on availability of required packages
    'TFModel', 'TFModule',
    'TorchModel', 'TorchModule',
    'SklearnModel', 'SklearnModule',
    'ONNXModel', 'ONNXModule',
]


try:
    from .tf_module import TFModel, TFModule
except ImportError:
    TFModel = None
    TFModule = None


try:
    from .torch_module import TorchModel, TorchModule
except ImportError:
    TorchModel = None
    TorchModule = None


try:
    from .sklearn_module import SklearnModel, SklearnModule
except ImportError:
    SklearnModel = None
    SklearnModule = None


try:
    from .onnx_module import ONNXModel, ONNXModule
except ImportError:
    ONNXModel = None
    ONNXModule = None


globals().update({
    "TFModel": TFModel,
    "TFModule": TFModule,
    "ONNXModel": ONNXModel,
    "ONNXModule": ONNXModule,
    "TorchModel": TorchModel,
    "TorchModule": TorchModule,
    "SklearnModel": SklearnModel,
    "SklearnModule": SklearnModule,
})
